import unittest

import numpy as np

from whatifact._widgets import _get_single_slider, _get_drop_down, _get_null_checkbox


class TestContinuous(unittest.TestCase):
    def setUp(self):
        self.arr1 = np.linspace(-5, 5, 100)
        self.arr2 = np.arange(10)

        mask = np.random.binomial(p=0.1, n=1, size=self.arr1.shape)
        self.arr3 = self.arr1.copy()
        self.arr3[mask] = np.nan

        self.v1 = dict(
            id="id", caption="caption", min=1, max=10, value=3, step=1, null=True
        )
        self.v2 = dict(
            id="id", caption="caption", min=1, max=10, value=3, step=1, null=False
        )

    def test_get_single_slider_allow_nulls(self):
        slider = _get_single_slider(v=self.v1)
        html = slider.get_html_string()

        self.assertIn(f'data-min="{self.v1["min"]}"', html)
        self.assertIn(f'data-max="{self.v1["max"]}"', html)
        self.assertIn(f"checkbox", html)

    def test_get_single_slider_allow_nulls_false(self):
        slider = _get_single_slider(v=self.v2)
        html = slider.get_html_string()

        self.assertIn(f'data-min="{self.v2["min"]}"', html)
        self.assertIn(f'data-max="{self.v2["max"]}"', html)
        self.assertNotIn(f"checkbox", html)

    def test_get_null_checkbox(self):
        checkbox = _get_null_checkbox(v=self.v1)
        html = checkbox.get_html_string()

        self.assertIn("id_null", html)


class TestCategorical(unittest.TestCase):
    def setUp(self):
        self.v = dict(
            id="id", caption="caption", options=["a", "b", "c"], value="b", null=False
        )

    def test_get_drop_down(self):
        selector = _get_drop_down(self.v)
        html = selector.get_html_string()
        for option in self.v["options"]:
            self.assertIn(f'value="{option}"', html)


if __name__ == "__main__":
    unittest.main()



"""
Data preprocessing functions
"""

from typing import List, Dict, Union

from shiny import ui
import pandas as pd

from whatifact._widgets import (
    _get_drop_down,
    _get_single_slider,
    _disable_slider,
    _enable_slider,
)
from whatifact._inference import _get_sliders_params, _get_select_list_params


def _get_variables_and_widgets(
    df,
    continuous_features: List[str],
    categorical_features: List[str],
    feature_settings: Dict[str, Dict[str, Union[int, float]]] = None,
):
    variables = dict()  # pylint: disable=use-dict-literal
    widgets = dict()  # pylint: disable=use-dict-literal
    feature_settings = feature_settings or dict()  # pylint: disable=use-dict-literal

    for j, col in enumerate(df.columns, start=1):
        var_id = f"var{j}"
        var_dict = {"id": var_id, "caption": col}

        if col in continuous_features:
            var_dict.update(
                **_get_sliders_params(
                    df[col], col_dict=feature_settings.get(col, dict())
                )  # pylint: disable=use-dict-literal
            )

        elif col in categorical_features:
            var_dict.update(
                **_get_select_list_params(
                    df[col],
                    col_dict=feature_settings.get(
                        col, dict()
                    ),  # pylint: disable=use-dict-literal
                )
            )

        variables.update({var_id: var_dict})
        widgets.update(
            {
                var_id: (
                    _get_single_slider(v=var_dict)
                    if var_dict["type"] == "continuous"
                    else _get_drop_down(v=var_dict)
                )
            }
        )

    widgets.update({"footer": ui.card_footer(" ")})

    return variables, widgets


def _update_values(df, sample, variables):
    row = df.loc[[sample]]
    for var_id, v in variables.items():
        col = v["caption"]
        new_value = row.loc[sample, col]

        if v["type"] == "continuous":
            if pd.isna(new_value):
                _disable_slider(var_id)
                ui.update_checkbox(var_id + "_null", value=False)
            elif v["null"]:
                _enable_slider(var_id, variables[var_id], org_value=new_value)
                ui.update_checkbox(var_id + "_null", value=True)
            else:
                ui.update_slider(var_id, value=float(new_value))

        elif v["type"] == "categorical":
            ui.update_select(
                id=v["id"], selected=str(new_value) if ~pd.isna(new_value) else ""
            )
            if v["null"]:
                ui.update_checkbox(var_id + "_null", value=~pd.isna(new_value))



"""
server function for shiny App
"""

from shiny import ui, reactive, Inputs, Outputs, Session, render
import numpy as np
import pandas as pd

from whatifact._utilities import _disable_slider, _enable_slider
from whatifact._utilities import _update_values


def whatifact_server(
    inputs: Inputs,
    output: Outputs,
    session: Session,
    variables: dict,
    sample_id_type,
    df: pd.DataFrame,
    clf: object,
):
    """
    server function for shiny UI app.
    Includes all UI logic
    """
    null_checkboxes = [
        n
        for n in inputs._map.keys()
        if n.startswith("var")
        and n.endswith("_null")  # pylint: disable=protected-access
    ]

    def get_null_checkboxes_dict():
        with reactive.isolate():
            null_dict = {n[:-5]: inputs[n]() for n in null_checkboxes}
            return null_dict

    def get_params():
        null_checkboxes_dict = get_null_checkboxes_dict()
        params_dict = {}

        for var_id, v in variables.items():
            col = v["caption"]
            params_dict.update(
                {
                    col: (
                        [inputs[var_id]()]
                        if ~null_checkboxes_dict.get(var_id, False)
                        else [np.nan]
                    )
                }
            )

            params = pd.DataFrame(params_dict)

            if v["type"] == "categorical" and inputs[var_id]() == "":
                if null_checkboxes_dict[var_id]:
                    ui.update_checkbox(var_id + "_null", value=False)

                if (df[col].dtype == "category") and (params.loc[0, col] != ""):
                    cat_dtype = df[col].cat.categories.dtype
                    params[col] = params[col].astype(cat_dtype)

        params = params.astype(df.dtypes.replace({int: "float64"}))
        return params

    def get_active_sample_id():
        with reactive.isolate():
            new_sample_id = sample_id_type(inputs.sample_id())
            return new_sample_id

    state = reactive.Value()
    state.dict = get_null_checkboxes_dict()

    @reactive.effect
    def checkbox_change():
        null_checkboxes_dict = {n[:-5]: inputs[n]() for n in null_checkboxes}

        for var_id, null_value in null_checkboxes_dict.items():

            if null_value == state.dict[var_id]:  # if checkbox is unchanged, continue
                continue

            if null_value:
                if variables[var_id]["type"] == "continuous":
                    _enable_slider(
                        var_id,
                        var_dict=variables[var_id],
                        org_value=df.loc[
                            get_active_sample_id(), variables[var_id]["caption"]
                        ],
                    )
                else:
                    ui.update_select(
                        var_id,
                        selected=df.loc[
                            get_active_sample_id(), variables[var_id]["caption"]
                        ],
                    )

            else:
                if variables[var_id]["type"] == "continuous":
                    _disable_slider(var_id)
                else:
                    ui.update_select(var_id, selected="")

        # refresh state dict
        state.dict = get_null_checkboxes_dict()

    @reactive.effect
    def sync_sample_id():
        sample = sample_id_type(inputs["sample_id"]())
        _update_values(
            df=df,
            sample=sample,
            variables=variables,
        )

    @reactive.effect
    @reactive.event(inputs.revert)
    def revert_changes():
        sample = sample_id_type(inputs["sample_id"]())
        _update_values(
            df=df,
            sample=sample,
            variables=variables,
        )

    @reactive.calc
    @render.text()
    def calc_pred() -> str:
        params = get_params()

        try:
            prediction = clf.predict_proba(params)[:, 1].item()
            assert isinstance(prediction, float)

        except Exception:  # pylint: disable=broad-exception-caught
            exit_app()

        prediction = np.round(prediction, 3)
        return prediction

    @reactive.effect
    @reactive.event(inputs.exit)
    async def exit_app():
        await session.close()



import unittest
import warnings

import pandas as pd
import numpy as np
from lightgbm import LGBMClassifier
from sklearn.linear_model import LogisticRegression

from whatifact import whatifact


class TestWhatifact(unittest.TestCase):
    def setUp(self):
        n = 500
        self.df = pd.DataFrame(
            {
                "sample_id": np.random.choice(10000, size=n, replace=False),
                "Age": pd.Series(np.random.randint(25, 80, size=n)),
                "Height": pd.Series(np.random.randint(159, 191, size=n)),
                "Sex": pd.Series(
                    np.random.choice(["M", "F"], size=n, replace=True)
                ).astype("category"),
                "Hb": pd.Series(np.random.normal(loc=12, size=n)),
                "Diabetic": pd.Series(np.random.binomial(n=2, p=0.25, size=n)).astype(
                    "category"
                ),
            }
        )

        # preparing data for Logisitic regression
        self.df_numerical = self.df.copy()
        self.df_numerical["Sex"] = (
            self.df_numerical["Sex"].map({"M": 0, "F": 1}).astype(int)
        )
        self.df_numerical["Diabetic"] = self.df_numerical["Diabetic"].astype(int)

        # preparing outcomes
        outcome = (
            self.df["Age"] / self.df["Height"]
            + (self.df["Hb"] * self.df["Diabetic"].astype(int))
        ) * self.df["Sex"].apply(lambda x: 1 if x == "M" else 1.5).astype(int)
        prediction_threshold = np.quantile(outcome, q=0.75)
        self.labels = (outcome > prediction_threshold).astype(int)

    def test_with_lgbm(self):
        clf = LGBMClassifier(verbose=-1).fit(
            self.df.drop("sample_id", axis=1), self.labels
        )
        _ = whatifact(
            df=self.df,
            clf=clf,
            sample_id="sample_id",
            run_application=False,
        )

    def test_with_lgbm_with_missing_values(self):
        mask = np.random.binomial(n=1, p=0.05, size=self.df.shape).astype(bool)
        df_array = self.df.values
        df_array[mask] = np.nan
        df_missing = pd.DataFrame(df_array, columns=self.df.columns)
        df_missing["sample_id"] = self.df[
            "sample_id"
        ]  # sample_id should not have missing values
        df_missing = df_missing.astype(
            {
                "Age": float,
                "Height": float,
                "Sex": "category",
                "Hb": float,
                "Diabetic": "category",
            }
        )

        self.assertGreater(
            df_missing.isna().sum().sum(), 0
        )  # assert df_missing really has missing values

        clf = LGBMClassifier(verbose=-1).fit(
            df_missing.drop("sample_id", axis=1), self.labels
        )
        _ = whatifact(
            df=df_missing,
            clf=clf,
            sample_id="sample_id",
            run_application=False,
        )

    def test_with_logistic_regression(self):
        warnings.filterwarnings("ignore")
        clf = LogisticRegression(verbose=0).fit(
            self.df_numerical.drop("sample_id", axis=1), self.labels
        )
        _ = whatifact(
            df=self.df_numerical,
            clf=clf,
            sample_id="sample_id",
            run_application=False,
        )


if __name__ == "__main__":
    unittest.main()



"""
Main whatifact function and Shiny code
"""

from pathlib import Path

from shiny import App, Inputs, Outputs, Session, ui, run_app

from whatifact._widgets import _get_card_header
from whatifact._utilities import _get_variables_and_widgets
from whatifact._inference import _prepare_everything
from whatifact._server import whatifact_server

CSS_FILE = Path(__file__).parent / "resources" / "style.css"


def whatifact(
    df,
    clf,
    sample_id: str = None,
    feature_settings: dict = None,
    run_application: bool = True,
):  # pylint: disable=too-many-arguments
    """
    Creates a UI for counterfactual explorations, allowing users to modify input features
    and observe changes in model predictions.

    Parameters:
    -----------
    df : pandas.DataFrame
        The dataset containing the input features.
    clf : object
        A trained classification model that supports prediction and has a predict_proba method.
    sample_id : str, optional
        The identifier for a specific sample in the dataset (default is None).
    feature_settings : dict, optional
        A dictionary specifying constraints or custom settings for feature modifications.
        Will be inferenced if None.
    run_application : bool, optional
        Whether to launch the UI application immediately (default is True), or return an App object.
    """

    # assert data, prepare date
    (df, sample_id_type, continuous_features, categorical_features) = (
        _prepare_everything(
            df=df,
            clf=clf,
            sample_id=sample_id,
            feature_settings=feature_settings,
        )
    )

    # build widgets
    variables, widgets = _get_variables_and_widgets(
        df=df,
        continuous_features=continuous_features,
        categorical_features=categorical_features,
        feature_settings=feature_settings,
    )

    # Build UI
    app_ui = ui.page_fluid(
        ui.include_css(CSS_FILE),
        ui.card(
            _get_card_header(df, sample_id),
            ui.card_body(ui.div(*widgets.values(), class_="scrollable-card")),
        ),
    )

    def app_server(inputs: Inputs, outputs: Outputs, session: Session):
        whatifact_server(inputs, outputs, session, variables, sample_id_type, df, clf)

    app = App(app_ui, app_server)

    if run_application:
        run_app(app, launch_browser=True)

    return app



"""
Assertions and data inference functions
"""

from typing import Dict, Union, List, Tuple
from collections.abc import Callable

import pandas as pd
import numpy as np


def _prepare_everything(  # pylint: disable=too-many-arguments
    df: pd.DataFrame,
    clf: object,
    sample_id: str,
    feature_settings: Dict[str, Dict[str, Union[int, float, bool]]],
):

    # assert clf
    assert "predict_proba" in dir(clf), "clf must have a predict_proba method"
    assert (
        len(
            set(["feature_names_in_", "feature_name_", "feature_names"]).intersection(
                dir(clf)
            )
        )
        > 0
    ), "Could not find classifier's features"

    if "feature_names_in_" in dir(clf):  # logistic regression
        clf_features = clf.feature_names_in_
    elif "feature_name_" in dir(clf):  # LGBM
        clf_features = clf.feature_name_
    elif "feature_names" in dir(clf):  # XGboost
        clf_features = clf.feature_names
    else:
        clf_features = df.columns

    df = df.copy()

    # prepare sample_id
    df, sample_id_type = _handle_sample_id(df=df, sample_id=sample_id)

    # remove un-needed features
    df = df[clf_features]

    # infer continuous and categorical features
    if feature_settings:
        predefined_continuous = [
            k for k, v in feature_settings.items() if v.get("type") == "continuous"
        ]
        predefined_categorical = [
            k for k, v in feature_settings.items() if v.get("type") == "categorical"
        ]
    else:
        predefined_continuous = {}
        predefined_categorical = {}

    continuous_features = list(
        (set(_infer_continuous_features(df)) - set(predefined_categorical)).union(
            predefined_continuous
        )
    )
    print(f"Selected continuous features: {continuous_features}")

    categorical_features = list(
        (set(_infer_categorical_features(df)) - set(predefined_continuous)).union(
            predefined_categorical
        )
    )
    print(f"Selected categorical features: {categorical_features}")

    df = _assert_features(df, continuous_features, categorical_features)

    # assert feature_settings
    if feature_settings:
        _assert_feature_settings(df, feature_settings, continuous_features)

    return (df, sample_id_type, continuous_features, categorical_features)


def _handle_sample_id(
    df: pd.DataFrame, sample_id: str = None
) -> Tuple[pd.DataFrame, Callable, bool]:
    if sample_id:
        df.set_index(sample_id, inplace=True)

    # assert df's index is of a single type and has unique values
    assert len({type(c) for c in df.index}) == 1, "index must be of a single type"
    assert len(set(df.index)) == df.shape[0], "indices must be unique"

    # define index type
    sample_id_type = type(df.index[0])

    return df, sample_id_type


def _infer_continuous_features(df: pd.DataFrame) -> list:
    numerics = ["int16", "int32", "int64", "float16", "float32", "float64"]
    continuous_features = list(df.select_dtypes(include=numerics).columns)
    continuous_features = [col for col in continuous_features if df[col].nunique() > 5]
    return continuous_features


def _infer_categorical_features(df: pd.DataFrame) -> list:
    categorical_features = [
        col
        for col in df.columns
        if df[col].dtype == "object"
        or df[col].dtype.name == "category"
        or df[col].nunique() <= 5
    ]
    return categorical_features


def _assert_features(
    df: pd.DataFrame, continuous_features: list, categorical_features: list
) -> pd.DataFrame:

    # assertions
    assert (
        len(set(continuous_features).intersection(categorical_features)) == 0
    ), f"Continuous and categorical features overlap: {list(set(continuous_features).intersection(categorical_features))}"
    assert (
        len(set(continuous_features) - set(df.columns)) == 0
    ), f"Some of continuous_features are not in dataframe: {list(set(continuous_features) - set(df.columns))}"
    assert (
        len(set(categorical_features) - set(df.columns)) == 0
    ), f"Some of categorical_features are not in dataframe: {list(set(categorical_features) - set(df.columns))}"
    assert all(
        pd.api.types.is_numeric_dtype(df[col]) for col in continuous_features
    ), "continuous_features include non-numeric columns"
    assert all(
        df[col].dtype == "object"
        or df[col].dtype.name == "category"
        or df[col].nunique() <= 10
        for col in categorical_features
    ), "categorical_features has non-categorical features ('object', 'category', or >10 unique values)"

    # remove columns from data frame that are not required
    redundant_features = (
        set(df.columns) - set(continuous_features) - set(categorical_features)
    )
    df.drop(list(redundant_features), axis=1, inplace=True)

    return df


def _assert_feature_settings(
    df: pd.DataFrame,
    feature_settings: Dict[str, Dict[str, Union[int, float]]],
    continuous_features: List[str],
) -> None:
    for k, v in feature_settings.items():
        arr = df[k]
        if k in df.columns and k in continuous_features:
            if "min" in v.keys():
                assert (
                    v["min"] <= arr.min()
                ), f"'min' slider value smaller than minimal value in {k}"
            if "max" in v.keys():
                assert (
                    v["max"] >= arr.max()
                ), f"'max' slider value greater than maximal value in {k}"
        if v.get("null") is False:
            assert (
                arr.isna().sum() == 0
            ), f"'null' was defined False for feature `{k}`, but column has NaN values"


def _calculate_min_max_value(arr: np.array, func: str, decimals: int):
    min_max = (np.min, -1) if func == "min" else (np.max, 1)
    arr = np.array(arr)
    arr = arr[~np.isnan(arr)]
    min_max_value = min_max[0](arr) + (np.std(arr) * min_max[1])
    return round(min_max_value * 1.0, decimals).astype(arr.dtype).item()


def _calculate_step_size(arr: pd.Series, decimals: int):
    arr = np.array(arr)
    arr = arr[~np.isnan(arr)]
    step_size = np.mean(np.diff(np.sort(np.unique(arr))))
    return (
        np.max([round(step_size, decimals), round(0.1, decimals)])
        .astype(arr.dtype)
        .item()
    )


def _get_sliders_params(arr, col_dict=None):
    col_dict = col_dict or dict()  # pylint: disable=use-dict-literal
    arr = np.array(arr)
    decimals = col_dict.get("decimals", 1)
    return {
        "type": "continuous",
        "min": col_dict.get(
            "min", _calculate_min_max_value(arr, func="min", decimals=decimals)
        ),
        "max": col_dict.get(
            "max", _calculate_min_max_value(arr, func="max", decimals=decimals)
        ),
        "step": col_dict.get("step", _calculate_step_size(arr, decimals)),
        "value": round(arr[0], decimals).astype(arr.dtype).item(),
        "null": col_dict.get("null", any(pd.isna(arr))),
    }


def _get_select_list_params(arr, col_dict=None):
    col_dict = col_dict or dict()  # pylint: disable=use-dict-literal
    options = arr.dropna().sort_values().astype(str).unique().tolist()
    allow_nulls = col_dict.get("null", any(pd.isna(arr)))
    value = arr.iloc[0]
    value = "" if pd.isna(value) else str(value)

    return {
        "type": "categorical",
        "options": [""] + options if allow_nulls else options,
        "value": value,
        "null": allow_nulls,
    }



from setuptools import setup, find_packages
import tomli

with open('pyproject.toml', 'rb') as f:
    version = tomli.load(f)['project']['version']

setup(
    name='whatifact',
    version=version,
    packages=find_packages() + ['whatifact.resources'],
    test_suite='tests',
    url='http://github.com/urigott/whatifact',
    author_email='urigott@gmail.com',
    license='MIT',
    install_requires=[
        "numpy>=1.26.0",
        "pandas>=1.1.0",
        "shiny>=1.0.0",
    ]

)


"""
Shiny widgets are defined here
"""

import numpy as np
import pandas as pd
from shiny import ui

def _get_null_checkbox(v, value=False):
    return ui.column(1, ui.input_checkbox(id=v["id"] + "_null", label="", value=value))

def _get_single_slider(v: dict):
    row = ui.row()

    if v["null"]:
        initial_value = ~np.isnan(v["value"])
        row.append(_get_null_checkbox(v, value=initial_value))
    else:
        row.append(ui.column(1, ""))

    row.append(
        ui.column(
            9,
            ui.input_slider(
                id=v["id"],
                label=v["caption"],
                min=v["min"],
                max=v["max"],
                value=float(v["value"]),
                step=v["step"],
            ),
        )
    )
    return row


def _get_drop_down(v):
    row = ui.row()

    if v["null"]:
        initial_value = v["value"] != ""
        row.append(_get_null_checkbox(v, value=initial_value))
    else:
        row.append(ui.column(1, ""))

    row.append(
        ui.column(
            9,
            ui.input_select(
                v["id"], v["caption"], choices=v["options"], selected=v["value"]
            ),
        )
    )

    return row


def _disable_slider(var_id):
    ui.update_slider(var_id, min=0, max=0)


def _enable_slider(var_id, var_dict, org_value):
    value = max(var_dict["min"], org_value)
    ui.update_slider(
        var_id, min=var_dict["min"], max=var_dict["max"], value=float(value)
    )


def _get_card_header(df: pd.DataFrame, sample_id: str = None):
    card_header = ui.card_header(
        ui.div(
            ui.input_select(
                id="sample_id",
                label=sample_id or "Sample ID",
                choices=list(df.index.astype(str)),
            ),
            class_="center-select",
        ),
        ui.div(
            "Predicted probability",
            ui.row(ui.output_text_verbatim("calc_pred")),
            class_="prediction-box",
        ),
        # ui.input_action_button(id='revert', label='⟳')
        class_="fixed-header",
    )
    return card_header



from .whatifact import whatifact
import tomli

with open('pyproject.toml', 'rb') as f:
    version = tomli.load(f)['project']['version']

__all__ = ["whatifact"]
__author__ = "Uri Gottlieb <urigott@gmail.com>"
__version__ = version



import unittest
from string import ascii_letters

import pandas as pd
import numpy as np


from whatifact._inference import (
    _calculate_step_size,
    _calculate_min_max_value,
    _handle_sample_id,
    _get_sliders_params,
    _infer_continuous_features,
    _infer_categorical_features,
    _assert_features,
    _assert_feature_settings,
)


class TestAssertions(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                "col1": [1, 2, 3] * 10,  # categorical, nunique() < 5
                "col2": list(ascii_letters[:30]),  # categorical, type object
                "col3": [i for i in range(10)]
                * 3,  # categorical, will be defined below
                "col4": np.arange(30),  # continuous, int
                "col5": np.random.normal(size=30),  # continuous, float
            }
        )

        self.df["col3"] = self.df["col3"].astype("category")
        self.categorical_features = ["col1", "col2", "col3"]
        self.continuous_features = ["col4", "col5"]

    def test_infer_continuous_features(self):
        continuous_features = set(_infer_continuous_features(self.df))
        self.assertEqual(continuous_features, {"col4", "col5"})

    def test_infer_categorical_features(self):
        categorical_features = set(_infer_categorical_features(self.df))
        self.assertEqual(categorical_features, {"col1", "col2", "col3"})

    def test_assert_features_pass(self):
        _ = _assert_features(
            df=self.df,
            categorical_features=self.categorical_features,
            continuous_features=self.continuous_features,
        )

    def test_assert_features_fail_overlap(self):
        with self.assertRaises(AssertionError) as e:
            _ = _assert_features(
                df=self.df,
                categorical_features=self.categorical_features,
                continuous_features=self.continuous_features
                + [self.categorical_features[0]],
            )
        self.assertIn("Continuous and categorical features overlap", str(e.exception))

    def test_assert_features_fail_unavailable_continuous_features(self):
        with self.assertRaises(AssertionError) as e:
            _ = _assert_features(
                df=self.df,
                categorical_features=self.categorical_features,
                continuous_features=self.continuous_features + ["col9"],
            )
        self.assertIn(
            "Some of continuous_features are not in dataframe", str(e.exception)
        )

    def test_assert_features_fail_unavailable_categorical_features(self):
        with self.assertRaises(AssertionError) as e:
            _ = _assert_features(
                df=self.df,
                categorical_features=self.categorical_features + ["col9"],
                continuous_features=self.continuous_features,
            )
        self.assertIn(
            "Some of categorical_features are not in dataframe", str(e.exception)
        )

    def test_assert_features_fail_wrong_continuous_features(self):
        with self.assertRaises(AssertionError) as e:
            _ = _assert_features(
                df=self.df,
                categorical_features=self.categorical_features[:2],
                continuous_features=self.continuous_features
                + [self.categorical_features[-1]],
            )
        self.assertIn(
            "continuous_features include non-numeric columns", str(e.exception)
        )

    def test_assert_features_fail_wrong_categorical_features(self):
        with self.assertRaises(AssertionError) as e:
            _ = _assert_features(
                df=self.df,
                categorical_features=self.categorical_features
                + [self.continuous_features[-1]],
                continuous_features=self.continuous_features[:1],
            )
        self.assertIn(
            "categorical_features has non-categorical features", str(e.exception)
        )

    def test_assert_features_drop_columns(self):
        df_redundant = self.df.copy().assign(col6=np.arange(30))

        df_asserted = _assert_features(
            df_redundant,
            categorical_features=self.categorical_features,
            continuous_features=self.continuous_features,
        )
        self.assertEqual(set(self.df.columns), set(df_asserted.columns))

    def test_assert_feature_settings_pass(self):
        feature_settings = {"col4": {"min": -5, "max": 100, "step": 2, "decimals": 3}}
        _assert_feature_settings(
            df=self.df,
            feature_settings=feature_settings,
            continuous_features=self.continuous_features,
        )

    def test_assert_feature_settings_min_too_high(self):
        feature_settings = {"col4": {"min": 1, "max": 100, "step": 2, "decimals": 3}}
        with self.assertRaises(AssertionError) as e:
            _assert_feature_settings(
                df=self.df,
                feature_settings=feature_settings,
                continuous_features=self.continuous_features,
            )
        self.assertIn("'min' slider value smaller", str(e.exception))

    def test_assert_feature_settings_max_too_low(self):
        feature_settings = {"col4": {"min": -1, "max": 20, "step": 2, "decimals": 3}}
        with self.assertRaises(AssertionError) as e:
            _assert_feature_settings(
                df=self.df,
                feature_settings=feature_settings,
                continuous_features=self.continuous_features,
            )
        self.assertIn("'max' slider value greater", str(e.exception))


class TestHandleSampleID(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                "sample_id": np.arange(30) + 3000,
                "col1": [1, 2, 3] * 10,  # categorical, nunique() < 5
                "col2": list(ascii_letters[:30]),  # categorical, type object
                "col3": [i for i in range(10)]
                * 3,  # categorical, will be defined below
                "col4": np.arange(30),  # continuous, int
                "col5": np.random.normal(size=30),  # continuous, float
            }
        )
        self.categorical_features = ["col1", "col2", "col3"]
        self.continuous_features = ["col4", "col5"]

    def test_handle_sample_id_pass(self):
        df, sample_id_type = _handle_sample_id(df=self.df, sample_id="sample_id")

        self.assertEqual(len(df.columns), 5)
        self.assertEqual(df.index.name, "sample_id")
        self.assertEqual(sample_id_type("3"), 3)

    def test_handle_sample_id_no_sample_id(self):
        df, sample_id_type = _handle_sample_id(df=self.df)

        self.assertEqual(len(df.columns), 6)
        self.assertEqual(df.index.name, None)
        self.assertEqual(sample_id_type("3"), 3)

    def test_handle_sample_id_string_sample_id(self):
        df, sample_id_type = _handle_sample_id(df=self.df, sample_id="col2")

        self.assertEqual(len(df.columns), 5)
        self.assertEqual(df.index.name, "col2")
        self.assertEqual(sample_id_type(3), "3")

    def test_handle_sample_id_no_multiple_sample_id_dtype(self):
        df = self.df.copy()
        df["new_sample_id"] = [i for i in range(15)] + list(ascii_letters[:15])

        with self.assertRaises(AssertionError) as e:
            df, sample_id_type = _handle_sample_id(df=df, sample_id="new_sample_id")
        self.assertIn("index must be of a single type", str(e.exception))

    def test_handle_sample_id_non_unique_sample_id(self):
        df = self.df.copy()
        df.loc[0, "sample_id"] = df.loc[1, "sample_id"]

        with self.assertRaises(AssertionError) as e:
            df, sample_id_type = _handle_sample_id(df=df, sample_id="sample_id")
        self.assertIn("indices must be unique", str(e.exception))


class TestInference(unittest.TestCase):
    def setUp(self):
        self.arr1 = np.linspace(-5, 5, 100)
        self.arr2 = np.arange(10)

        mask = np.random.binomial(p=0.1, n=1, size=self.arr1.shape)
        self.arr3 = self.arr1.copy()
        self.arr3[mask] = np.nan

        self.v = dict(id="id", caption="caption", min=1, max=10, value=3, step=1)

    def test_calculate_min_max_value(self):
        r = _calculate_min_max_value(self.arr1, func="min", decimals=1)
        self.assertLess(r, self.arr1.min())

        r = _calculate_min_max_value(self.arr1, func="max", decimals=1)
        self.assertGreater(r, self.arr1.max())
        self.assertIsInstance(r, float)

        r = _calculate_min_max_value(self.arr2, func="max", decimals=1)
        self.assertGreater(r, self.arr1.max())
        self.assertIsInstance(r, int)

    def test_calculate_min_max_value_with_missing(self):
        r = _calculate_min_max_value(self.arr3, func="min", decimals=1)
        self.assertLess(r, np.nanmin(self.arr3))
        self.assertIsInstance(r, float)

    def test_calculate_step_size(self):
        r = _calculate_step_size(self.arr2, decimals=1)
        self.assertEqual(r, 1)

    def test_calculate_step_size_with_missing(self):
        r = _calculate_step_size(self.arr3, decimals=1)
        self.assertIsInstance(r, float)

    def test_calculate_step_size(self):
        r = _calculate_step_size(self.arr1, decimals=3)
        self.assertIsInstance(r, float)

    def test_get_sliders_params(self):
        r = _get_sliders_params(self.arr2)

        self.assertIsInstance(r, dict)
        self.assertEqual(set(r.keys()), {"min", "max", "step", "value", "null", "type"})
        self.assertEqual(r["value"], 0.0)


if __name__ == "__main__":
    unittest.main()



import unittest
from string import ascii_letters

import numpy as np
import pandas as pd

from whatifact._utilities import _get_variables_and_widgets, _update_values


class TestPreprocessing(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {
                "sample_id": np.arange(30) + 3000,
                "col1": [1, 2, 3] * 10,  # categorical, nunique() < 5
                "col2": list(ascii_letters[:30]),  # categorical, type object
                "col3": [i for i in range(10)]
                * 3,  # categorical, will be defined below
                "col4": np.arange(30),  # continuous, int
                "col5": np.random.normal(size=30),  # continuous, float
            }
        )

        self.categorical_features = ["col1", "col2", "col3"]
        self.continuous_features = ["col4", "col5"]

    def test_get_variables_and_widgets(self):
        variables, widgets = _get_variables_and_widgets(
            df=self.df.set_index("sample_id"),
            continuous_features=self.continuous_features,
            categorical_features=self.categorical_features,
        )

        self.assertEqual(len(variables), len(widgets) - 1)
        self.assertTrue(all([isinstance(v, str) for v in variables]))
        self.assertTrue(all([isinstance(v, dict) for v in variables.values()]))

        self.assertTrue(all(["id" in v.keys() for v in variables.values()]))
        self.assertTrue(all(["caption" in v.keys() for v in variables.values()]))
        self.assertTrue(all(["type" in v.keys() for v in variables.values()]))
        self.assertTrue(
            all(
                [
                    "options" in v.keys()
                    for v in variables.values()
                    if v["type"] == "categorical"
                ]
            )
        )
        self.assertTrue(
            all(
                [
                    "value" in v.keys()
                    for v in variables.values()
                    if v["type"] == "continuous"
                ]
            )
        )

    def test_update_values(self):
        pass


if __name__ == "__main__":
    unittest.main()
